
EvaluateMonomial := function( mono, imgs )
    local w, i;
    if IsInt(mono[1]) then 
        w := imgs[mono[1]];
        for i in [2..Length(mono)] do
            w := w * imgs[mono[i]];
        od;
        return w;
    elif IsList(mono[1]) then 
        return EvaluateMonomial(mono[1], imgs)*EvaluateMonomial(mono[2],imgs);
    fi;
end;

EvaluateWord := function( word, imgs )
    local w, i;
    w := word[2] * EvaluateMonomial( word[1], imgs );
    for i in [3,5..Length(word)-1] do
        w := w + word[i+1] * EvaluateMonomial( word[i], imgs );
    od;
    return w; 
end;
   
EvaluatePresentation := function( M, rels, imgs )
    local i, a;
    for i in [1..Length(rels)] do
        a := EvaluateWord( rels[i], imgs ); 
        if a <> Zero(M) then return false; fi;
    od;
    return true;
end;

IsIsomorphismByPres := function( L, M, imgs )
    if not EvaluatePresentation( M, L!.rels, imgs ) then return false; fi;
    if Dimension(Subalgebra(M,imgs)) < Dimension(M) then return false; fi;
    return true;
end;
    
LieRandomIsomorphismTest := function( L, M, r )
    local n, f, i, g;

    # set up L
    AddPresentationOfLieAlg(L);
    n := Length(L!.gens);

    # set up M
    f := L!.type;
    if not IsBound(M!.elem) then M!.elem := ElementsByType(M, f); fi;
    if Length(M!.elem) <> L!.nrtyp then return false; fi;

    # loop 
    for i in [1..r*Dimension(L)] do
        Print("trial number ",i,"\n");
        g := List([1..n], x -> Random(M!.elem));
        if IsIsomorphismByPres( L, M, g ) then return true; fi; 
    od;
 
    # not found
    return fail;
end;
 
LieIsomorphismTest := function(L, M)
    local n, f, i, j;

    # set up L
    AddPresentationOfLieAlg(L);
    n := Length(L!.gens);

    # set up M
    f := L!.type;
    if not IsBound(M!.elem) then M!.elem := ElementsByType(M, f); fi;
    if Length(M!.elem) <> L!.nrtyp then return false; fi;

    # loop
    for i in [0..Length(M!.elem)^n-1] do
        j := CoefficientsQadic( i, Length(M!.elem) ); 
        j := List( j, x -> M!.elem[x+1]);
        if IsIsomorphismByPres(L, M, j) then return true; fi;
    od;

    return false;
end;

MatrixImage := function( L, hom )
    return List(Basis(L), x -> Coefficients(Basis(L), Image(hom,x)));
end;

LieRandomAutomorphismGroup := function( L, e, r )
    local G, A, g, n, i, h, aut;

    # set up presentation
    AddPresentationOfLieAlg(L);
    g := L!.gens * Basis(L);
    n := Length(L!.gens);

    # set up autos
    G := GL( Dimension(L), LeftActingDomain(L) );
    A := SubgroupNC( G, [] );

    for i in [1..r] do
        h := List([1..n], x -> Random(e));
        Print("RAG: trial ",i," \n");
        if IsIsomorphismByPres( L, L, h ) then 
            aut := AlgebraHomomorphismByImagesNC(L, L, g, h);
            aut := MatrixImage(L,aut);
            if not aut in A then  
                A := ClosureSubgroupNC( A, aut );
                Print("RAG: got group of size ",Size(A),"\n");
            fi;
        fi;
    od;
    return A;
end;

LieAutGroupByElms := function(L, e)
    local G, A, c, i, aut, g, f, n;

    # set up 
    AddPresentationOfLieAlg(L);
    g := L!.gens * Basis(L);
    n := Length(L!.gens);

    # set up autos
    G := GL( Dimension(L), LeftActingDomain(L) );
    A := SubgroupNC( G, [] );

    # set up elements
    c := Cartesian(List([1..n], x -> e));
    Print("RAG: loop over ",Length(c)," options \n");

    for i in [1..Length(c)] do
        if IsIsomorphismByPres( L, L, c[i] ) then 
            aut := AlgebraHomomorphismByImagesNC(L, L, g, c[i]);
            aut := MatrixImage(L,aut);
            if not aut in A then  
                A := ClosureSubgroupNC( A, aut );
                Print("RAG: got group of size ",Size(A),"\n");
            fi;
        fi;
    od;
    return A;
end;

LieAutGroup := function(L)
    local s, d, I, f, e, g;
    s := SamplePolsAndElms(L);

    # take those of minimal degree
    d := Minimum(List(s.pols, Degree));
    I := Filtered([1..Length(s.pols)], x -> Degree(s.pols[x])=d);
    s.pols := s.pols{I}; s.elms := s.elms{I};

    # choose that with the smallest number of elements
    d := Minimum(List(s.elms, Length));
    I := Filtered([1..Length(s.elms)], x -> Length(s.elms[x])=d);
    f := s.pols[Random(I)];
    L!.type := f;
    Print("got type ",f,"\n");

    e := ElementsByType(L, f);
    Print("got ",Length(e)," elements \n");

    g := FindGensInSet(L, e);
    L!.gens := List(g, x -> Coefficients(Basis(L),x));
    Print("got ",Length(g)," gens \n");
    L!.rels := PresentationByGens(L, L!.gens);

    return LieAutGroupByElms(L, e);
end;
 
GenericSubspaces := function(L)
    local e, U, V;
    e := NilpotentElements(L);
    U := Subalgebra(L,e);
    e := SolvableElements(L);
    V := Subalgebra(L,e);
    return [U,V];
end;
        
TrySamples := function(L)
    local l, A, i, g, s;
    l := GenericSubspaces(L);
    Print(" subalgs are ",Dimension(l[1])," ",Dimension(l[2])," ",
            Dimension(Intersection(l[1],l[2])),"\n");
    s := SamplePolsAndElms(L);
    s := SampleElms(L, s);
    A := List([1..Length(s.elms)], x -> false);
    Print("consider ",List(s.elms, Length)," options \n");
    for i in [1..Length(s.elms)] do
        g := FindGensInSet(L, s.elms[i]);
        L!.gens := List(g, x -> Coefficients(Basis(L),x));
        L!.type := s.pols[i];
        L!.rels := PresentationByGens(L, L!.gens);
        A[i] := LieAutGroupByElms(L, s.elms[i]);
        Print(i," done \n\n");
    od;
    return A;
end;


